@@ -24,6 +24,24 @@ typedef signed __int32    int32_t;
 #endif
 
 
+#if defined(ANNOYLIB_USE_AVX512)
+#define AVX_INFO "Using 512-bit AVX instructions"
+#elif defined(ANNOYLIB_USE_AVX128)
+#define AVX_INFO "Using 128-bit AVX instructions"
+#else
+#define AVX_INFO "Not using AVX instructions"
+#endif
+
+#if defined(_MSC_VER)
+#define COMPILER_INFO "Compiled using MSC"
+#elif defined(__GNUC__)
+#define COMPILER_INFO "Compiled on GCC"
+#else
+#define COMPILER_INFO "Compiled on unknown platform"
+#endif
+
+#define ANNOY_DOC (COMPILER_INFO ". " AVX_INFO ".")
+
 #if PY_MAJOR_VERSION >= 3
 #define IS_PY3K
 #endif
@@ -36,8 +54,15 @@ typedef signed __int32    int32_t;
     #define PyInt_FromLong PyLong_FromLong 
 #endif
 
+using namespace Annoy;
 
-template class AnnoyIndexInterface<int32_t, float>;
+#ifdef ANNOYLIB_MULTITHREADED_BUILD
+  typedef AnnoyIndexMultiThreadedBuildPolicy AnnoyIndexThreadedBuildPolicy;
+#else
+  typedef AnnoyIndexSingleThreadedBuildPolicy AnnoyIndexThreadedBuildPolicy;
+#endif
+
+template class Annoy::AnnoyIndexInterface<int32_t, float>;
 
 class HammingWrapper : public AnnoyIndexInterface<int32_t, float> {
   // Wrapper class for Hamming distance, using composition.
@@ -45,7 +70,7 @@ class HammingWrapper : public AnnoyIndexInterface<int32_t, float> {
   // This is questionable from a performance point of view. Should reconsider this solution.
 private:
   int32_t _f_external, _f_internal;
-  AnnoyIndex<int32_t, uint64_t, Hamming, Kiss64Random> _index;
+  AnnoyIndex<int32_t, uint64_t, Hamming, Kiss64Random, AnnoyIndexThreadedBuildPolicy> _index;
   void _pack(const float* src, uint64_t* dst) const {
     for (int32_t i = 0; i < _f_internal; i++) {
       dst[i] = 0;
@@ -61,18 +86,18 @@ private:
   };
 public:
   HammingWrapper(int f) : _f_external(f), _f_internal((f + 63) / 64), _index((f + 63) / 64) {};
-  void add_item(int32_t item, const float* w) {
+  bool add_item(int32_t item, const float* w, char**error) {
     vector<uint64_t> w_internal(_f_internal, 0);
     _pack(w, &w_internal[0]);
-    _index.add_item(item, &w_internal[0]);
+    return _index.add_item(item, &w_internal[0], error);
   };
-  void build(int q) { _index.build(q); };
-  void unbuild() { _index.unbuild(); };
-  bool save(const char* filename, bool prefault) { return _index.save(filename, prefault); };
+  bool build(int q, int n_threads, char** error) { return _index.build(q, n_threads, error); };
+  bool unbuild(char** error) { return _index.unbuild(error); };
+  bool save(const char* filename, bool prefault, char** error) { return _index.save(filename, prefault, error); };
   void unload() { _index.unload(); };
-  bool load(const char* filename, bool prefault) { return _index.load(filename, prefault); };
+  bool load(const char* filename, bool prefault, char** error) { return _index.load(filename, prefault, error); };
   float get_distance(int32_t i, int32_t j) const { return _index.get_distance(i, j); };
-  void get_nns_by_item(int32_t item, size_t n, size_t search_k, vector<int32_t>* result, vector<float>* distances) const {
+  void get_nns_by_item(int32_t item, size_t n, int search_k, vector<int32_t>* result, vector<float>* distances) const {
     if (distances) {
       vector<uint64_t> distances_internal;
       _index.get_nns_by_item(item, n, search_k, result, &distances_internal);
@@ -81,7 +106,7 @@ public:
       _index.get_nns_by_item(item, n, search_k, result, NULL);
     }
   };
-  void get_nns_by_vector(const float* w, size_t n, size_t search_k, vector<int32_t>* result, vector<float>* distances) const {
+  void get_nns_by_vector(const float* w, size_t n, int search_k, vector<int32_t>* result, vector<float>* distances) const {
     vector<uint64_t> w_internal(_f_internal, 0);
     _pack(w, &w_internal[0]);
     if (distances) {
@@ -100,8 +125,8 @@ public:
     _index.get_item(item, &v_internal[0]);
     _unpack(&v_internal[0], v);
   };
-  void set_seed(int q) { _index.set_seed(q); };
-  bool on_disk_build(const char* filename) { return _index.on_disk_build(filename); };
+  void set_seed(uint64_t q) { _index.set_seed(q); };
+  bool on_disk_build(const char* filename, char** error) { return _index.on_disk_build(filename, error); };
 };
 
 // annoy python object
@@ -123,16 +148,21 @@ py_an_new(PyTypeObject *type, PyObject *args, PyObject *kwargs) {
   static char const * kwlist[] = {"f", "metric", NULL};
   if (!PyArg_ParseTupleAndKeywords(args, kwargs, "i|s", (char**)kwlist, &self->f, &metric))
     return NULL;
-  if (!metric || !strcmp(metric, "angular")) {
-   self->ptr = new AnnoyIndex<int32_t, float, Angular, Kiss64Random>(self->f);
+  if (!metric) {
+    // This keeps coming up, see #368 etc
+    PyErr_WarnEx(PyExc_FutureWarning, "The default argument for metric will be removed "
+		 "in future version of Annoy. Please pass metric='angular' explicitly.", 1);
+    self->ptr = new AnnoyIndex<int32_t, float, Angular, Kiss64Random, AnnoyIndexThreadedBuildPolicy>(self->f);
+  } else if (!strcmp(metric, "angular")) {
+   self->ptr = new AnnoyIndex<int32_t, float, Angular, Kiss64Random, AnnoyIndexThreadedBuildPolicy>(self->f);
   } else if (!strcmp(metric, "euclidean")) {
-    self->ptr = new AnnoyIndex<int32_t, float, Euclidean, Kiss64Random>(self->f);
+    self->ptr = new AnnoyIndex<int32_t, float, Euclidean, Kiss64Random, AnnoyIndexThreadedBuildPolicy>(self->f);
   } else if (!strcmp(metric, "manhattan")) {
-    self->ptr = new AnnoyIndex<int32_t, float, Manhattan, Kiss64Random>(self->f);
+    self->ptr = new AnnoyIndex<int32_t, float, Manhattan, Kiss64Random, AnnoyIndexThreadedBuildPolicy>(self->f);
   } else if (!strcmp(metric, "hamming")) {
     self->ptr = new HammingWrapper(self->f);
   } else if (!strcmp(metric, "dot")) {
-    self->ptr = new AnnoyIndex<int32_t, float, DotProduct, Kiss64Random>(self->f);
+    self->ptr = new AnnoyIndex<int32_t, float, DotProduct, Kiss64Random, AnnoyIndexThreadedBuildPolicy>(self->f);
   } else {
     PyErr_SetString(PyExc_ValueError, "No such metric");
     return NULL;
@@ -170,8 +200,7 @@ static PyMemberDef py_annoy_members[] = {
 
 static PyObject *
 py_an_load(py_annoy *self, PyObject *args, PyObject *kwargs) {
-  char* filename;
-  bool res = false;
+  char *filename, *error;
   bool prefault = false;
   if (!self->ptr) 
     return NULL;
@@ -179,10 +208,9 @@ py_an_load(py_annoy *self, PyObject *args, PyObject *kwargs) {
   if (!PyArg_ParseTupleAndKeywords(args, kwargs, "s|b", (char**)kwlist, &filename, &prefault))
     return NULL;
 
-  res = self->ptr->load(filename, prefault);
-
-  if (!res) {
-    PyErr_SetFromErrno(PyExc_IOError);
+  if (!self->ptr->load(filename, prefault, &error)) {
+    PyErr_SetString(PyExc_IOError, error);
+    free(error);
     return NULL;
   }
   Py_RETURN_TRUE;
@@ -191,8 +219,7 @@ py_an_load(py_annoy *self, PyObject *args, PyObject *kwargs) {
 
 static PyObject *
 py_an_save(py_annoy *self, PyObject *args, PyObject *kwargs) {
-  char *filename;
-  bool res = false;
+  char *filename, *error;
   bool prefault = false;
   if (!self->ptr) 
     return NULL;
@@ -200,10 +227,9 @@ py_an_save(py_annoy *self, PyObject *args, PyObject *kwargs) {
   if (!PyArg_ParseTupleAndKeywords(args, kwargs, "s|b", (char**)kwlist, &filename, &prefault))
     return NULL;
 
-  res = self->ptr->save(filename, prefault);
-
-  if (!res) {
-    PyErr_SetFromErrno(PyExc_IOError);
+  if (!self->ptr->save(filename, prefault, &error)) {
+    PyErr_SetString(PyExc_IOError, error);
+    free(error);
     return NULL;
   }
   Py_RETURN_TRUE;
@@ -212,21 +238,48 @@ py_an_save(py_annoy *self, PyObject *args, PyObject *kwargs) {
 
 PyObject*
 get_nns_to_python(const vector<int32_t>& result, const vector<float>& distances, int include_distances) {
-  PyObject* l = PyList_New(result.size());
-  for (size_t i = 0; i < result.size(); i++)
-    PyList_SetItem(l, i, PyInt_FromLong(result[i]));
+  PyObject* l = NULL;
+  PyObject* d = NULL;
+  PyObject* t = NULL;
+
+  if ((l = PyList_New(result.size())) == NULL) {
+    goto error;
+  }
+  for (size_t i = 0; i < result.size(); i++) {
+    PyObject* res = PyInt_FromLong(result[i]);
+    if (res == NULL) {
+      goto error;
+    }
+    PyList_SetItem(l, i, res);
+  }
   if (!include_distances)
     return l;
 
-  PyObject* d = PyList_New(distances.size());
-  for (size_t i = 0; i < distances.size(); i++)
-    PyList_SetItem(d, i, PyFloat_FromDouble(distances[i]));
+  if ((d = PyList_New(distances.size())) == NULL) {
+    goto error;
+  }
 
-  PyObject* t = PyTuple_New(2);
-  PyTuple_SetItem(t, 0, l);
-  PyTuple_SetItem(t, 1, d);
+  for (size_t i = 0; i < distances.size(); i++) {
+    PyObject* dist = PyFloat_FromDouble(distances[i]);
+    if (dist == NULL) {
+      goto error;
+    }
+    PyList_SetItem(d, i, dist);
+  }
+
+  if ((t = PyTuple_Pack(2, l, d)) == NULL) {
+    goto error;
+  }
+  Py_XDECREF(l);
+  Py_XDECREF(d);
 
   return t;
+
+  error:
+    Py_XDECREF(l);
+    Py_XDECREF(d);
+    Py_XDECREF(t);
+    return NULL;
 }
 
 
@@ -269,24 +322,31 @@ py_an_get_nns_by_item(py_annoy *self, PyObject *args, PyObject *kwargs) {
 
 bool
 convert_list_to_vector(PyObject* v, int f, vector<float>* w) {
-  if (PyObject_Size(v) == -1) {
-    char buf[256];
-    snprintf(buf, 256, "Expected an iterable, got an object of type \"%s\"", v->ob_type->tp_name);
-    PyErr_SetString(PyExc_ValueError, buf);
+  Py_ssize_t length = PyObject_Size(v);
+  if (length == -1) {
     return false;
   }
-  if (PyObject_Size(v) != f) {
-    char buf[128];
-    snprintf(buf, 128, "Vector has wrong length (expected %d, got %ld)", f, PyObject_Size(v));
-    PyErr_SetString(PyExc_IndexError, buf);
+  if (length != f) {
+    PyErr_Format(PyExc_IndexError, "Vector has wrong length (expected %d, got %ld)", f, length);
     return false;
   }
+
   for (int z = 0; z < f; z++) {
     PyObject *key = PyInt_FromLong(z);
+    if (key == NULL) {
+      return false;
+    }
     PyObject *pf = PyObject_GetItem(v, key);
-    (*w)[z] = PyFloat_AsDouble(pf);
     Py_DECREF(key);
+    if (pf == NULL) {
+      return false;
+    }
+    double value = PyFloat_AsDouble(pf);
     Py_DECREF(pf);
+    if (value == -1.0 && PyErr_Occurred()) {
+      return false;
+    }
+    (*w)[z] = value;
   }
   return true;
 }
@@ -333,11 +393,22 @@ py_an_get_item_vector(py_annoy *self, PyObject *args) {
   vector<float> v(self->f);
   self->ptr->get_item(item, &v[0]);
   PyObject* l = PyList_New(self->f);
+  if (l == NULL) {
+    return NULL;
+  }
   for (int z = 0; z < self->f; z++) {
-    PyList_SetItem(l, z, PyFloat_FromDouble(v[z]));
+    PyObject* dist = PyFloat_FromDouble(v[z]);
+    if (dist == NULL) {
+      goto error;
+    }
+    PyList_SetItem(l, z, dist);
   }
 
   return l;
+
+  error:
+    Py_XDECREF(l);
+    return NULL;
 }
 
 
@@ -359,25 +430,28 @@ py_an_add_item(py_annoy *self, PyObject *args, PyObject* kwargs) {
   if (!convert_list_to_vector(v, self->f, &w)) {
     return NULL;
   }
-  self->ptr->add_item(item, &w[0]);
+  char* error;
+  if (!self->ptr->add_item(item, &w[0], &error)) {
+    PyErr_SetString(PyExc_Exception, error);
+    free(error);
+    return NULL;
+  }
 
   Py_RETURN_NONE;
 }
 
 static PyObject *
 py_an_on_disk_build(py_annoy *self, PyObject *args, PyObject *kwargs) {
-  char *filename;
-  bool res = false;
+  char *filename, *error;
   if (!self->ptr)
     return NULL;
   static char const * kwlist[] = {"fn", NULL};
   if (!PyArg_ParseTupleAndKeywords(args, kwargs, "s", (char**)kwlist, &filename))
     return NULL;
 
-  res = self->ptr->on_disk_build(filename);
-
-  if (!res) {
-    PyErr_SetFromErrno(PyExc_IOError);
+  if (!self->ptr->on_disk_build(filename, &error)) {
+    PyErr_SetString(PyExc_IOError, error);
+    free(error);
     return NULL;
   }
   Py_RETURN_TRUE;
@@ -386,15 +460,23 @@ py_an_on_disk_build(py_annoy *self, PyObject *args, PyObject *kwargs) {
 static PyObject *
 py_an_build(py_annoy *self, PyObject *args, PyObject *kwargs) {
   int q;
+  int n_jobs = -1;
   if (!self->ptr) 
     return NULL;
-  static char const * kwlist[] = {"n_trees", NULL};
-  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "i", (char**)kwlist, &q))
+  static char const * kwlist[] = {"n_trees", "n_jobs", NULL};
+  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "i|i", (char**)kwlist, &q, &n_jobs))
     return NULL;
 
+  bool res;
+  char* error;
   Py_BEGIN_ALLOW_THREADS;
-  self->ptr->build(q);
+  res = self->ptr->build(q, n_jobs, &error);
   Py_END_ALLOW_THREADS;
+  if (!res) {
+    PyErr_SetString(PyExc_Exception, error);
+    free(error);
+    return NULL;
+  }
 
   Py_RETURN_TRUE;
 }
@@ -404,10 +486,13 @@ static PyObject *
 py_an_unbuild(py_annoy *self) {
   if (!self->ptr) 
     return NULL;
-  
-  Py_BEGIN_ALLOW_THREADS;
-  self->ptr->unbuild();
-  Py_END_ALLOW_THREADS;
+
+  char* error;
+  if (!self->ptr->unbuild(&error)) {
+    PyErr_SetString(PyExc_Exception, error);
+    free(error);
+    return NULL;
+  }
 
   Py_RETURN_TRUE;
 }
@@ -495,7 +580,7 @@ static PyMethodDef AnnoyMethods[] = {
   {"get_item_vector",(PyCFunction)py_an_get_item_vector, METH_VARARGS, "Returns the vector for item `i` that was previously added."},
   {"add_item",(PyCFunction)py_an_add_item, METH_VARARGS | METH_KEYWORDS, "Adds item `i` (any nonnegative integer) with vector `v`.\n\nNote that it will allocate memory for `max(i)+1` items."},
   {"on_disk_build",(PyCFunction)py_an_on_disk_build, METH_VARARGS | METH_KEYWORDS, "Build will be performed with storage on disk instead of RAM."},
-  {"build",(PyCFunction)py_an_build, METH_VARARGS | METH_KEYWORDS, "Builds a forest of `n_trees` trees.\n\nMore trees give higher precision when querying. After calling `build`,\nno more items can be added."},
+  {"build",(PyCFunction)py_an_build, METH_VARARGS | METH_KEYWORDS, "Builds a forest of `n_trees` trees.\n\nMore trees give higher precision when querying. After calling `build`,\nno more items can be added. `n_jobs` specifies the number of threads used to build the trees. `n_jobs=-1` uses all available CPU cores."},
   {"unbuild",(PyCFunction)py_an_unbuild, METH_NOARGS, "Unbuilds the tree in order to allows adding new items.\n\nbuild() has to be called again afterwards in order to\nrun queries."},
   {"unload",(PyCFunction)py_an_unload, METH_NOARGS, "Unloads an index from disk."},
   {"get_distance",(PyCFunction)py_an_get_distance, METH_VARARGS, "Returns the distance between items `i` and `j`."},
@@ -528,7 +613,7 @@ static PyTypeObject PyAnnoyType = {
   0,                      /*tp_setattro*/
   0,                      /*tp_as_buffer*/
   Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE, /*tp_flags*/
-  "annoy objects",        /* tp_doc */
+  ANNOY_DOC,              /* tp_doc */
   0,                      /* tp_traverse */
   0,                      /* tp_clear */
   0,                      /* tp_richcompare */
@@ -556,7 +641,7 @@ static PyMethodDef module_methods[] = {
   static struct PyModuleDef moduledef = {
     PyModuleDef_HEAD_INIT,
     "annoylib",          /* m_name */
-    "",                  /* m_doc */
+    ANNOY_DOC,           /* m_doc */
     -1,                  /* m_size */
     module_methods,      /* m_methods */
     NULL,                /* m_reload */
